#include "interface.h"
#include "graphplan.h"
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#include <limits.h>

char *bb_return_val_names[5] = { "Unsat", "Sat", "Timeout", "Failure", "Simplified" };


int bb_printflag;

float bb_maxsec;
int bb_maxit;

long bb_maxticks;
long bb_starttick;
struct tms bb_buffer;
long bb_current_it;
int bb_muststop;

void bb_init_limitp(float maxsec, int maxit)
{
    bb_maxticks = (long)(CLK_TCK * maxsec);
    bb_maxit = maxit;
    bb_current_it = 0;
    times(&bb_buffer);
    bb_starttick = bb_buffer.tms_utime + bb_buffer.tms_stime;
    bb_muststop = FALSE;
}

int bb_limitp(int rate)
{
    if (bb_muststop) return TRUE;
    bb_current_it++;
    if (bb_maxit && (bb_current_it >= bb_maxit)) {
	bb_muststop = TRUE;
	return TRUE;
    }
    if (bb_maxticks && bb_current_it % rate == 0){
	times(&bb_buffer);
	if ((bb_buffer.tms_utime + bb_buffer.tms_stime - bb_starttick) >= bb_maxticks){
	    bb_muststop = TRUE;
	    return TRUE;
	}
    }
    return FALSE;
}

/************************************************/
/*  lists                                       */
/************************************************/

#define Appendi(list, i) {\
    if (list ## _len >= list ## _buflen) Stretchi(&list, &list ## _buflen);\
    list[ list ## _len ++ ] = i; }

#define Appendp(list, p) {\
    if (list ## _len >= list ## _buflen) Stretchp(&list, &list ## _buflen);\
    list[ list ## _len ++ ] = p; }

void Stretchi(int ** list, int * buflen)
{
    if ((*buflen)<=0){
	*list = (int *) malloc((size_t)(sizeof(int)*1024));
	*buflen = 1024;
    }
    else {
	*list = (int *) realloc(*list, (size_t)(sizeof(int)*(*buflen)*2));
	*buflen *= 2;
    }
}

void Stretchp(vertex_t ** list, int * buflen)
{
    if ((*buflen)<=0){
	*list = (vertex_t *) malloc((size_t)(sizeof(void *)*1024));
	*buflen = 1024;
    }
    else {
	*list = (vertex_t *) realloc(*list, (size_t)(sizeof(void *)*(*buflen)*2));
	*buflen *= 2;
    }
}

/************************************************/
/*  converting graph to wff                     */
/************************************************/

int bb_numvar;
int bb_numclause;

int * bb_wff;
int bb_wff_len;
int bb_wff_buflen = 0;

int * bb_soln;
int bb_soln_len;
int bb_soln_buflen = 0;

vertex_t * bb_prop2vertex;
int bb_prop2vertex_len;
int bb_prop2vertex_buflen = 0;


void bb_generate_mutex(hashtable_t t, int stamp)
{
  vertex_t vert;
  edgelist_t edge;

  get_next(t, 0); /* get ready */
  while((vert = get_next(t, 1)) != NULL) {
    if (vert->needed == 0) continue;
    if ((edge = vert->exclusive) != NULL) {
	for(; edge; edge = edge->next){
	    if (edge->endpt->needed == 0) continue;
	    bb_numclause++;
	    if (bb_printflag & PrintLit) fprintf(STDDATA, "( NOT %s-%d    NOT %s-%d )\n", vert->name, stamp, edge->endpt->name, stamp); 
	    Appendi(bb_wff, -(vert->prop));
	    Appendi(bb_wff, -(edge->endpt->prop));
	    Appendi(bb_wff, 0);
	}
    }
  }
}


void bb_generate_frame(hashtable_t t, int stamp, int simplified)
{
  vertex_t vert;
  vertex_t op;
  edgelist_t edge;

  get_next(t, 0); /* get ready */
  while((vert = get_next(t, 1)) != NULL) {
    if (vert->needed == 0) continue;

    /* note that at maxtime you know for sure the precondition
       of axiom holds, so you can eliminate it */

    if (simplified){
	if (bb_printflag & PrintLit) fprintf(STDDATA, "( ");
    }
    else {
	if (bb_printflag & PrintLit) fprintf(STDDATA, "( NOT %s-%d    ", vert->name, stamp);
    }

    if (! simplified) Appendi(bb_wff, -(vert->prop));

    for (edge = vert->in_edges; edge; edge = edge->next){
	op = edge->endpt;
	if (op->needed == 0) continue;
	if (bb_printflag & PrintLit) fprintf(STDDATA, " %s-%d ", op->name, stamp);
	Appendi(bb_wff, op->prop);
    }
    bb_numclause++;
    if (bb_printflag & PrintLit) fprintf(STDDATA, " )\n");
    Appendi(bb_wff, 0);
  }
}
  
void bb_generate_precond(hashtable_t t, int stamp)
{
  vertex_t vert;
  vertex_t fact;
  edgelist_t edge;

  get_next(t, 0); /* get ready */
  while((vert = get_next(t, 1)) != NULL) {
    if (vert->needed == 0) continue;
    for (edge = vert->in_edges; edge; edge = edge->next){
	fact = edge->endpt;
	if (fact->needed == 0) {
	    fprintf(STDMSG,"***ERROR**** unneeded precondition for %s-%d\n", vert->name, stamp);
	}
	bb_numclause++;
	if (bb_printflag & PrintLit) fprintf(STDDATA, "( NOT %s-%d    %s-%d )\n", vert->name, stamp, fact->name, stamp-1);
	Appendi(bb_wff, - (vert->prop));
	Appendi(bb_wff, fact->prop);
	Appendi(bb_wff, 0);
    }
  }
}

void bb_assign_prop_table(hashtable_t t, int stamp)
{
    vertex_t vert;

    get_next(t, 0); /* get ready */
    while((vert = get_next(t, 1)) != NULL) {
	if (vert->needed == 0) continue;
	vert->prop = ++ bb_numvar;
	if (bb_printflag & PrintMap)
	    fprintf(STDDATA, "a %d %s-%d\n", vert->prop, vert->name, stamp);
	Appendp(bb_prop2vertex, vert);
    }
}

void bb_assign_prop(int len, int num_goals)
{
    int i;

    bb_numvar = 0;
    bb_prop2vertex_len = 0;
    Appendp(bb_prop2vertex, NULL);

    for(i=0; i < len; ++i) {
	bb_assign_prop_table(fact_table[i], i);
	bb_assign_prop_table(op_table[i], i+1);
    }
    bb_assign_prop_table(fact_table[len], len);
    bb_soln_len = 0;
    Appendi(bb_soln, 0);
    for (i=1; i<=bb_numvar; i++)
	Appendi(bb_soln, 0);
    fprintf(STDMSG, "bb_soln array length %d out of %d\n", bb_soln_len, bb_soln_buflen);
}



void bb_generate_wff(int len, int num_goals)
{
  int i;
  vertex_t v;

  remove_unneeded_vertices(len);
  bb_assign_prop(len, num_goals);
  fprintf(STDMSG, "%d variables\n", bb_numvar);
  bb_numclause = 0;
  bb_wff_len = 0;

  if (bb_printflag & PrintLit) fprintf(STDMSG, "Begin literal wff\n");

  /* print initial facts */
  /* probably not necessary, but cannot hurt */
  get_next(fact_table[0],0);
  while((v = get_next(fact_table[0],1)) != NULL) {
      if (v->needed == 0) continue;
      if (v->is_true == 1){ /* should always hold, because of setup() */
	  bb_numclause++;
	  if (bb_printflag & PrintLit) fprintf(STDDATA, "( %s-%d )\n", v->name, 0);
	  Appendi(bb_wff, v->prop); 
	  Appendi(bb_wff, 0);
      }
  }

  /* print goals */
  /* also probably not necessary, due to special handling of frame axioms */
  for(i=0; i < num_goals; ++i) {
      bb_numclause++;
      if (goals_at[len][i]->needed == 0) continue;
      if (bb_printflag & PrintLit) fprintf(STDDATA, "( %s-%d )\n",goals_at[len][i]->name, len);
      Appendi(bb_wff, goals_at[len][i]->prop);
      Appendi(bb_wff, 0);
  }

  /* print mutex and backward chaining */
  for(i=0; i < len; ++i) {
      bb_generate_mutex(op_table[i], i+1);
      bb_generate_precond(op_table[i], i+1);

      bb_generate_mutex(fact_table[i+1], i+1);
      bb_generate_frame(fact_table[i+1], i+1, (i+1)==len);
  }
  if (bb_printflag & PrintLit) fprintf(STDMSG, "End literal wff\n");
  fprintf(STDMSG, "%d clauses\n", bb_numclause);
}


void bb_print_cnf(void)
{
    int i,j;

    fprintf(STDMSG, "Begin cnf wff\n");
    fprintf(STDDATA, "p cnf %d %d\n", bb_numvar, bb_numclause);
    i=0; j = -1;
    while (i<bb_numclause){
	do {
	    fprintf(STDDATA, " %d", bb_wff[++j]);
	} while (bb_wff[j] != 0);
	fprintf(STDDATA, "\n");
	i++;
    }
    fprintf(STDMSG, "End cnf wff\n");
}

/************************************************/
/*  main routine                                */
/************************************************/

void bb_graph2wff(int maxtime, int num_goals)
{
    fprintf(STDMSG,"Converting graph to wff\n");
    bb_generate_wff(maxtime, num_goals);
    if (bb_printflag & PrintCNF) bb_print_cnf();
    if (bb_printflag & PrintExit) exit(0);
}

/***********************************************/
/*  convert solution back to graph             */
/***********************************************/

void bb_soln2graph(void)
{
    int i;

    if (bb_printflag & PrintModel) fprintf(STDDATA, "Begin model\n");
    for (i=1; i<=bb_numvar; i++){
	bb_prop2vertex[i]->is_true = bb_soln[i];
	bb_prop2vertex[i]->used = bb_soln[i];
	if (bb_printflag & PrintModel){
	    fprintf(STDDATA, " %d", bb_soln[i] ? i : -i);
	    if (i % 10 == 0) fprintf(STDDATA, "\n");
	}
    }
    if (bb_printflag & PrintModel) fprintf(STDDATA, "\nEnd model\n");
}

/************************************************/
/*  solver specificiations                      */
/************************************************/

int bb_spec_len = 0;
bb_solver_spec bb_spec[20];

void bb_parse_solver_args(int i, int argc, char * argv[])
{
    int k, a;
    int (*p)(int, int, int *, int *, int, char * [] );

    k = 0;
    bb_spec_len = 0;
    i++;
    while (i < argc){
	if (strcmp(argv[i], "-maxsec")==0){
	    if (++i >= argc) do_error("command line args not in proper format");
	    sscanf(argv[i], "%f", &bb_maxsec);
	}
	else if (strcmp(argv[i], "-maxit")==0){
	    if (++i >= argc) do_error("command line args not in proper format");
	    sscanf(argv[i], "%d", &bb_maxit);
	}
	else {

	    k = bb_spec_len++;

	    bb_spec[k].solver = Anysat;
	    if (strcmp(argv[i], "graphplan")==0){
		p = NULL;
		bb_spec[k].solver = Graphplan;
	    }
	    else if (strcmp(argv[i], "compact")==0){
		bb_spec[k].solver = Compact;
		p = & bb_simplifier_compact;
	    }
	    else if (strcmp(argv[i], "satz")==0)
		p = & bb_satsolve_satz;
	    else if (strcmp(argv[i], "relsat")==0)
		p = & bb_satsolve_relsat;
	    else if (strcmp(argv[i], "walksat")==0)
		p = & bb_satsolve_walksat;
	    else do_error("command line args not in proper format");
	    
	    bb_spec[k].maxsec = bb_maxsec;
	    bb_spec[k].maxit = bb_maxit;
	    bb_spec[k].argv = &argv[i];
	    bb_spec[k].proc = p;

	    a = 1;
	    while ( ++i < argc && strcmp(argv[i], "-then")!=0 ) a++;
	    bb_spec[k].argc = a;
	}
	i++;
    }
}


void bb_print_solver_spec(void)
{
    int i,j;

    fprintf(STDMSG, "Begin solver specification\n");
    for (i=0; i<bb_spec_len; i++){
	fprintf(STDMSG,"    -maxint %8d   -maxsec %8f ", bb_spec[i].maxit, (double)bb_spec[i].maxsec);
	for (j=0; j<bb_spec[i].argc; j++){
	    fprintf(STDMSG, " %s", bb_spec[i].argv[j]);
	}
	fprintf(STDMSG, "\n");
    }
    fprintf(STDMSG, "End solver specification\n");
}



void bb_reparse_graphplan_options(int argc, char * argv[])
{
    int i,j;
    char * option;

    for(j = 1; j < argc; ++j) {
	if (strcmp(argv[j], "-O")==0 && j+1 < argc){
	    option = argv[++j];

	    do_subsets = 0;
	    do_irrel = 0;
	    do_buildup = 0;
	    do_greedy = 0;
	    do_lower = 0;
	    do_noexcl = 0;
	    for(i=0; option[i] != '\0'; ++i) {
		if (option[i] == 'S') {
		    do_subsets = 1; 
		    fprintf(STDMSG, "Graphplan option: S == subsets\n");
		    do_completeness = 0;}
		if (option[i] == 'I') {
		    fprintf(STDMSG, "Graphplan option: I == irrelevent\n");
		    do_irrel = 1; }
		if (option[i] == 'B') {
		    fprintf(STDMSG, "Graphplan option: B == buildup\n");
		    do_buildup = 1; }
		if (option[i] == 'G') {
		    fprintf(STDMSG, "Graphplan option: G == greedy\n");
		    do_greedy = 1; }
		if (option[i] == 'L') { 
		    fprintf(STDMSG, "Graphplan option: L == lower bound\n");
		    do_lower = 1; do_completeness = 0;  }
		if (option[i] == 'E') {
		    fprintf(STDMSG, "Graphplan option: E == no exclusiveness\n");
		    do_noexcl = 1; }
	    }
	}
	else {
	    fprintf(STDMSG, "Bad argument to graphplan\n");
	    exit(1);
	}
    }
}

void bb_graphplan_help(void)
{
    fprintf(STDMSG, "
Options for graphplan solver are:
     -O <option list> where
     I = look for irrelevants
     L = Lower bound time needed by counting steps
     B = Build up to goals
     E = Don't do mutual exclusions (for testing)
     S = examine subsets
");
}

char * bb_default_argstring[] = { 
    "-solver", "-maxsec", "30", "graphplan", "-then", 
    "compact", "-b", "4", "-then",
    "-maxsec", "0", "satz", "-cutoff", "20", "-restart", "10", "-then", 
    "satz", "-cutoff", "200", "-restart", "10"
};


char * bb_satz_argstring[] = { "-solver", "satz" };
char * bb_help_argv[] = { "dummy", "-help" };

void bb_default_solver_args(int helpmode)
{
    int i;

    if (helpmode){
	for (i=0; i<bb_spec_len; i++){
	    if (bb_spec[i].solver == Graphplan)
		bb_graphplan_help();
	    else
		(void)(* (bb_spec[i].proc))(0, 0, NULL, NULL, 2, bb_help_argv);
	}
	exit(1);
    }

    if (bb_spec_len == 0){

	if (bb_printflag){
	    bb_parse_solver_args(0, 2, bb_satz_argstring);
	    return;
	}

	bb_parse_solver_args(0, sizeof bb_default_argstring / sizeof bb_default_argstring[0], 
			     bb_default_argstring);
    }
    bb_print_solver_spec();
}


int bb_satsolve_relsat(int numvar, int numclause, int * wff, int * soln, 
		       int argc, char * argv[])
{
    fprintf(STDMSG, "Invoking solver rel_sat_rand\n");

    fprintf(STDMSG, "NOT YET IMPLEMENTED\n");

    
    return Timeout;
}


