/******************************************************************************/
/* satz-rand.c                                                                */
/* based on satz.c by Chu Min Li 1996                                         */
/* randomization by Bart Selman 1987                                          */
/* further modifications by Henry Kautz 1998                                  */
/******************************************************************************/

#include "../interface.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <limits.h>

#include <string.h>
#include <math.h>
#include <signal.h>
#include <string.h>

/* Can't find header files on Sun for srandom
   (not in stdlib.h!)
   - but live with warnings, because these
   declarations break on linux! 
long random(void);
int srandom( unsigned seed);
*/

#define VERSION "satz-rand-2.1"

typedef signed char my_type;
typedef unsigned char my_unsigned_type;

#define DYNAMIQUE_SIZE 256

/* #define WORD_LENGTH 50 */
#define WORD_LENGTH 256

#define TRUE 1
#define FALSE 0
#define NONE -1
#define TOOBIG -2

#define WEIGTH 5
#define T 10

/* Change these variables to enlarge problems that can be handled.
When tab_clause_size becomes 1000000 or greater, satz-rand becomes
unstable, and often "hangs".  There may be some arrays that are the
declared to be the wrong size and overflow. */

#define tab_variable_size  20000
#define tab_clause_size 500000


#define tab_variable_size_plus_1 tab_variable_size+1
#define tab_clause_size_plus_1 tab_clause_size+1
#define tab_literal_size 2*tab_variable_size
#define double_tab_clause_size 2*tab_clause_size
#define positive(literal) literal<NB_VAR
#define negative(literal) literal>=NB_VAR
#define get_var_from_lit(negative_literal) negative_literal-NB_VAR
#define RESOLVANT_LENGTH 3
#define RESOLVANT_SEARCH_THRESHOLD 5000
#define complement(lit1, lit2) ((lit1<lit2) ? lit2-lit1 == NB_VAR : lit1-lit2 == NB_VAR)

#define pop(stack) stack[--stack ## _fill_pointer]
#define push(item, stack) stack[stack ## _fill_pointer++] = item
#define unit_clause(clause) clause_length[clause] == 1

#define dolist(ppointer, pend, begin, end)                                  \
          for(ppointer = begin, pend = end; ppointer != pend; ppointer++)


#define empty_vector(pstack) pstack ## _fill_pointer == 0

#define satisfiable() (CLAUSE_STACK_fill_pointer == NB_CLAUSE)

#define NEGATIVE 0
#define POSITIVE 1
#define PASSIVE 0
#define ACTIVE 1
#define MAX_NODE_NUMBER 6000

#define OLD_CLAUSE_REDUNDANT 77
#define NEW_CLAUSE_REDUNDANT 7

#define SAVE(x)  { x ## _SAVE = x ; }
#define RESTORE(x)  { x = x ## _SAVE ; }
#define SAVEA(x,i)  { x ## _SAVE[i] = x[i] ; }
#define RESTOREA(x,i)  { x[i] = x ## _SAVE[i] ; }
#define SAVEV(x) { for (i=0; i<tab_variable_size; i++) { x ## _SAVE[i] = x[i] ; } }
#define RESTOREV(x) { for (i=0; i<tab_variable_size; i++) { x[i] = ## _SAVE[i] ; } }
#define SAVEC(x) { for (i=0; i<tab_clause_size; i++) { x ## _SAVE[i] = x[i] ; } }
#define RESTOREC(x) { for (i=0; i<tab_clause_size; i++) { x[i] = ## _SAVE[i] ; } }


long NB_SEARCH = 0; long NB_FIXED = 0;

struct node {
    int clause;
    struct node *next;
};

struct ressource {
     struct node *pnode;
     struct ressource *next;
};

int CLAUSE_NODES_pointer=MAX_NODE_NUMBER;

struct node *CLAUSE_NODES;
struct ressource *ressources=NULL;

struct node *allocate_node() {
   struct ressource *pressource;
   if (CLAUSE_NODES_pointer == MAX_NODE_NUMBER) {
      CLAUSE_NODES = (struct node *)malloc(MAX_NODE_NUMBER * sizeof(struct node));
      CLAUSE_NODES_pointer = 0;
      pressource= (struct ressource *)malloc(sizeof (struct ressource));
      pressource->next = ressources;
      ressources=pressource;
      pressource->pnode = CLAUSE_NODES;
   }  
   return &(CLAUSE_NODES[CLAUSE_NODES_pointer++]);
}

void free_ressources() {
   struct ressource *pressource;
   while (ressources != NULL) {
      free(ressources->pnode);
      pressource = ressources;
      ressources = ressources->next;
      free(pressource);
   }
}

struct node *node_neg_in[tab_variable_size];
struct node *node_pos_in[tab_variable_size];

int *neg_in[tab_variable_size];
int *pos_in[tab_variable_size];
my_unsigned_type neg_nb[tab_variable_size];
my_unsigned_type pos_nb[tab_variable_size];
my_type var_current_value[tab_variable_size];
my_type saved_soln[tab_variable_size];
my_type var_rest_value[tab_variable_size];
my_type var_state[tab_variable_size];

int saved_clause_stack[tab_variable_size];
int saved_managedclause_stack[tab_variable_size];
my_unsigned_type nb_neg_clause_of_length2[tab_variable_size];
my_unsigned_type nb_neg_clause_of_length3[tab_variable_size];
my_unsigned_type nb_pos_clause_of_length2[tab_variable_size];
my_unsigned_type nb_pos_clause_of_length3[tab_variable_size];

/* int reduce_if_positive[tab_variable_size][DYNAMIQUE_SIZE];
int reduce_if_negative[tab_variable_size][DYNAMIQUE_SIZE];
*/
int reduce_if_negative_nb[tab_variable_size];
int reduce_if_positive_nb[tab_variable_size];

int *sat[tab_clause_size];
my_type clause_state[tab_clause_size];
my_type clause_length[tab_clause_size];

/* my_type static_clause_length[tab_clause_size]; */

int VARIABLE_STACK_fill_pointer = 0;
int CLAUSE_STACK_fill_pointer = 0;
int UNITCLAUSE_STACK_fill_pointer = 0;
int MANAGEDCLAUSE_STACK_fill_pointer = 0;
int IMPLIED_LITS_fill_pointer=0;

int VARIABLE_STACK[tab_variable_size];
int CLAUSE_STACK[tab_clause_size];
int UNITCLAUSE_STACK[tab_clause_size];
int MANAGEDCLAUSE_STACK[double_tab_clause_size];
int IMPLIED_LITS[tab_variable_size];

int MY_VARIABLE_STACK[tab_variable_size];
/* int MY_CLAUSE_STACK[tab_clause_size]; */
int MY_UNITCLAUSE_STACK[tab_clause_size];
int MY_MANAGEDCLAUSE_STACK[double_tab_clause_size];

int TESTED_VAR_STACK[tab_variable_size];
int TESTED_VAR_STACK_fill_pointer=0;

int NB_VAR;
int NB_CLAUSE;
int INIT_NB_CLAUSE;
my_type R = 3;

long NB_UNIT=0, NB_MONO=0, NB_BRANCHE=0, NB_BACK = 0;

my_type MAYBE_SATISFIABLE = TRUE;

int MY_VARIABLE_STACK_fill_pointer = 0;
/* int MY_CLAUSE_STACK_fill_pointer = 0; */
int MY_UNITCLAUSE_STACK_fill_pointer = 0;
int MY_MANAGEDCLAUSE_STACK_fill_pointer = 0;

/* Randomization */
int seed;                      /* seed for random */
struct timeval tv;
struct timezone tzp;
int good_vars[tab_variable_size];
float factor = 0.4;           /* increment factor for chosen_var */
int cutoff = 1000000000;               /* increment factor for chosen_var */
int restart = 0;
int restart_max = 1;

void update_statistics_end_restart(int successful);


double elapsed_seconds(void)
{
    double answer;

    static struct tms prog_tms;
    static long prev_times = 0;

    (void) times(&prog_tms);

    answer = ((double)(((long)prog_tms.tms_utime)-prev_times))/((double) CLK_TCK);

    prev_times = (long) prog_tms.tms_utime;

    return answer;
}






void remove_clauses(int *clauses) {
   int clause;
   for(clause=*clauses; clause!=NONE; clause=*(++clauses)) {
     if (clause_state[clause] == ACTIVE) {
        clause_state[clause] = PASSIVE;
        push(clause, CLAUSE_STACK);
     }
  }
}

int manage_clauses(register int *clauses) {
   int clause;
   for(clause=*clauses; clause!=NONE; clause=*(++clauses)) {
      if (clause_state[clause] == ACTIVE) {
         switch (clause_length[clause]) {
            case 1: return FALSE;
            case 2: push(clause, UNITCLAUSE_STACK);
                    clause_length[clause]--; break;
            default: clause_length[clause]--;
                     push(clause, MANAGEDCLAUSE_STACK);
         }
      }
   }
   return TRUE;
}

void simple_manage_clauses(register int *clauses) {
   int clause;
   for(clause=*clauses; clause!=NONE; clause=*(++clauses)) { 
     if (clause_state[clause] == ACTIVE) {
         switch (clause_length[clause]) {
            case 2: push(clause, UNITCLAUSE_STACK);
                    clause_length[clause]--; break;
            default: clause_length[clause]--;
                     push(clause, MANAGEDCLAUSE_STACK);
         }
      }
   }
}

void my_simple_manage_clauses(register int *clauses) {
   int clause;
   for(clause=*clauses; clause!=NONE; clause=*(++clauses)) { 
      if (clause_state[clause] == ACTIVE) { 
         switch (clause_length[clause]) {
            case 2: push(clause, MY_UNITCLAUSE_STACK);
                    clause_length[clause]--; break;
            default: clause_length[clause]--; 
                     push(clause, MY_MANAGEDCLAUSE_STACK);
         }
      }
   }
}

int my_manage_clauses(register int *clauses) {
   int clause;
   for(clause=*clauses; clause!=NONE; clause=*(++clauses)) {
      if (clause_state[clause] == ACTIVE) { 
         switch (clause_length[clause]) {
            case 1: return FALSE;
            case 2: push(clause, MY_UNITCLAUSE_STACK);
                    clause_length[clause]--; break;
            default: clause_length[clause]--;
                     push(clause, MY_MANAGEDCLAUSE_STACK);
         }
      }
   }
   return TRUE;
}                                                             

void print_solution(int nb_var)
{
    int i;

    printf("\nBegin assignment\n");

    for (i=0; i<nb_var; i++) {
       if (saved_soln[i] == 1) 
          printf("%d ", i+1);
       else
          printf("%d ", 0-i-1);
       if ((i+1)%20 == 0) printf("\n");
    }
    printf("\n");
    printf("End assignment\n");
} 


void save_solution(int nb_var, int * soln)
{
    int i;
    
    for (i=0; i<nb_var; i++){
	saved_soln[i] = var_current_value[i];
	soln[i+1] = var_current_value[i];
    }
} 



void reset_context() {
   int i, tested_var;

   MY_UNITCLAUSE_STACK_fill_pointer = 0;

   for (i=0; i<MY_VARIABLE_STACK_fill_pointer; i++)
       var_state[MY_VARIABLE_STACK[i]] = ACTIVE;
   MY_VARIABLE_STACK_fill_pointer = 0;

   for (i=0; i<MY_MANAGEDCLAUSE_STACK_fill_pointer; i++)
         clause_length[MY_MANAGEDCLAUSE_STACK[i]]++;
   MY_MANAGEDCLAUSE_STACK_fill_pointer = 0;  
}

int* copy_clauses(struct node *node_in) {
    int j=0, *in; struct node *pnode;

    pnode=node_in;
    while (pnode != NULL) {
      j++; pnode = pnode->next;
    }
    in = (int *)malloc((j+1)*sizeof(int));
    j=0; pnode=node_in; 
    while (pnode != NULL) {
        in[j] = pnode->clause; j++; pnode = pnode->next; 
    }
    in[j] = NONE;
    return in;
}

void set_link(int clause) {
   int lit;
   int *lits;
   struct node *pnode1, *pnode2, *pnode;

   lits = sat[clause];
   for (lit=*lits; lit != NONE; lit=*(++lits)) { 
       pnode = allocate_node();
       pnode->clause = clause;
       if (positive(lit)) {
          pnode->next = node_pos_in[lit];
          node_pos_in[lit] = pnode;
       }
       else {
          pnode->next = node_neg_in[get_var_from_lit(lit)];
          node_neg_in[get_var_from_lit(lit)] = pnode;
       }
   }
}

my_type clause_length_SAVE[tab_clause_size];
my_type clause_state_SAVE[tab_clause_size];

int MY_MANAGEDCLAUSE_STACK_SAVE[double_tab_clause_size];
int MANAGEDCLAUSE_STACK_SAVE[double_tab_clause_size];

int CLAUSE_STACK_SAVE[tab_clause_size];
int saved_managedclause_stack_SAVE[tab_variable_size];
int saved_clause_stack_SAVE[tab_variable_size];
int UNITCLAUSE_STACK_SAVE[tab_clause_size];
int MY_UNITCLAUSE_STACK_SAVE[tab_clause_size];
int CLAUSE_STACK_fill_pointer_SAVE;
int MANAGEDCLAUSE_STACK_fill_pointer_SAVE;
int UNITCLAUSE_STACK_fill_pointer_SAVE;
int MY_UNITCLAUSE_STACK_fill_pointer_SAVE;
int MY_MANAGEDCLAUSE_STACK_fill_pointer_SAVE;

my_type var_state_SAVE[tab_variable_size];
my_type var_current_value_SAVE[tab_variable_size];
my_type var_rest_value_SAVE[tab_variable_size];
int reduce_if_positive_nb_SAVE[tab_variable_size];
int reduce_if_negative_nb_SAVE[tab_variable_size];
my_unsigned_type nb_neg_clause_of_length2_SAVE[tab_variable_size];
my_unsigned_type nb_neg_clause_of_length3_SAVE[tab_variable_size];
my_unsigned_type nb_pos_clause_of_length2_SAVE[tab_variable_size];
my_unsigned_type nb_pos_clause_of_length3_SAVE[tab_variable_size];
int IMPLIED_LITS_SAVE[tab_variable_size];
int VARIABLE_STACK_SAVE[tab_variable_size];
int TESTED_VAR_STACK_SAVE[tab_variable_size];
int IMPLIED_LITS_fill_pointer_SAVE;
int VARIABLE_STACK_fill_pointer_SAVE;
int TESTED_VAR_STACK_fill_pointer_SAVE;
long NB_FIXED_SAVE;
long NB_SEARCH_SAVE;
int NB_UNIT_SAVE;
int NB_MONO_SAVE;
int NB_BRANCHE_SAVE;
int NB_BACK_SAVE;

void save_everything()
{
    int i;

    for (i=0; i<double_tab_clause_size; i++){
	SAVEA(MY_MANAGEDCLAUSE_STACK, i);
	SAVEA(MANAGEDCLAUSE_STACK, i);
    }

    for (i=0; i<tab_clause_size; i++){
	SAVEA(clause_length, i);
	SAVEA(clause_state, i);
	SAVEA(CLAUSE_STACK, i);
	SAVEA(MANAGEDCLAUSE_STACK, i);
	SAVEA(UNITCLAUSE_STACK, i);
	SAVEA(MY_UNITCLAUSE_STACK, i);
	SAVEA(MY_MANAGEDCLAUSE_STACK, i);
    }
    SAVE(CLAUSE_STACK_fill_pointer);
    SAVE(MANAGEDCLAUSE_STACK_fill_pointer);
    SAVE(UNITCLAUSE_STACK_fill_pointer);
    SAVE(MY_UNITCLAUSE_STACK_fill_pointer);
    SAVE(MY_MANAGEDCLAUSE_STACK_fill_pointer);

    for (i=0; i<tab_variable_size; i++) {
	SAVEA(var_state,i);
	SAVEA(var_current_value,i);
	SAVEA(var_rest_value,i);
	SAVEA(reduce_if_positive_nb,i);
	SAVEA(reduce_if_negative_nb,i);
	SAVEA(nb_neg_clause_of_length2,i);
	SAVEA(nb_neg_clause_of_length3,i);
	SAVEA(nb_pos_clause_of_length2,i);
	SAVEA(nb_pos_clause_of_length3,i);
	SAVEA(IMPLIED_LITS, i);
	SAVEA(VARIABLE_STACK, i);
	SAVEA(TESTED_VAR_STACK, i);
    }
    SAVE(IMPLIED_LITS_fill_pointer);
    SAVE(VARIABLE_STACK_fill_pointer);
    SAVE(TESTED_VAR_STACK_fill_pointer);

    SAVE(NB_FIXED);
    SAVE(NB_SEARCH);
    SAVE(NB_UNIT);
    SAVE(NB_MONO);
    SAVE(NB_BRANCHE);
    SAVE(NB_BACK);
	
    /* does not appear to be used by the code anywhere */
    /* MY_CLAUSE_STACK_fill_pointer = 0; */

}

void restore_everything()
{
    int i;

    restart++;
    fprintf(STDMSG, "[%d] begin restart\n", restart);

    for (i=0; i<double_tab_clause_size; i++){
	RESTOREA(MY_MANAGEDCLAUSE_STACK, i);
	RESTOREA(MANAGEDCLAUSE_STACK, i);
    }

    for (i=0; i<tab_clause_size; i++){
	RESTOREA(clause_length, i);
	RESTOREA(clause_state, i);
	RESTOREA(CLAUSE_STACK, i);
	RESTOREA(MANAGEDCLAUSE_STACK, i);
	RESTOREA(UNITCLAUSE_STACK, i);
	RESTOREA(MY_UNITCLAUSE_STACK, i);
	RESTOREA(MY_MANAGEDCLAUSE_STACK, i);
    }
    RESTORE(CLAUSE_STACK_fill_pointer);
    RESTORE(MANAGEDCLAUSE_STACK_fill_pointer);
    RESTORE(UNITCLAUSE_STACK_fill_pointer);
    RESTORE(MY_UNITCLAUSE_STACK_fill_pointer);
    RESTORE(MY_MANAGEDCLAUSE_STACK_fill_pointer);

    for (i=0; i<tab_variable_size; i++) {
	RESTOREA(var_state,i);
	RESTOREA(var_current_value,i);
	RESTOREA(var_rest_value,i);
	RESTOREA(reduce_if_positive_nb,i);
	RESTOREA(reduce_if_negative_nb,i);
	RESTOREA(nb_neg_clause_of_length2,i);
	RESTOREA(nb_neg_clause_of_length3,i);
	RESTOREA(nb_pos_clause_of_length2,i);
	RESTOREA(nb_pos_clause_of_length3,i);
	RESTOREA(IMPLIED_LITS, i);
	RESTOREA(VARIABLE_STACK, i);
	RESTOREA(TESTED_VAR_STACK, i);
    }
    RESTORE(IMPLIED_LITS_fill_pointer);
    RESTORE(VARIABLE_STACK_fill_pointer);
    RESTORE(TESTED_VAR_STACK_fill_pointer);

    RESTORE(NB_FIXED);
    RESTORE(NB_SEARCH);
    RESTORE(NB_UNIT);
    RESTORE(NB_MONO);
    RESTORE(NB_BRANCHE);
    RESTORE(NB_BACK);
	
    /* does not appear to be used by the code anywhere */
    /* MY_CLAUSE_STACK_fill_pointer = 0; */

}

int backtracking() {
    int i, var, index, *plit, length, NB_CLAUSE1;
      
    UNITCLAUSE_STACK_fill_pointer = 0;
    NB_BACK++;

    if (bb_limitp(Satz_Rate)){
	return NONE;
    }

    if (NB_BACK > cutoff) {
	if (restart >= restart_max){
	    return NONE;
	}
	fprintf(STDMSG, "[%d] reached cutoff %d --- back to root\n", restart, cutoff);
	update_statistics_end_restart(FALSE);
	fflush(stdout);
	restore_everything();
	return TRUE;
    }

    do {
	var = pop(VARIABLE_STACK);
	if (var_rest_value[var] == NONE) 
	    var_state[var] = ACTIVE;
	else {
	    for (index = saved_clause_stack[var]; 
		 index < CLAUSE_STACK_fill_pointer;
		 index++)
		clause_state[CLAUSE_STACK[index]] = ACTIVE;
	    CLAUSE_STACK_fill_pointer = saved_clause_stack[var];

	    for (index = saved_managedclause_stack[var];
		 index < MANAGEDCLAUSE_STACK_fill_pointer;
		 index++)
		clause_length[MANAGEDCLAUSE_STACK[index]]++;
	    MANAGEDCLAUSE_STACK_fill_pointer = saved_managedclause_stack[var];

	    var_current_value[var] = var_rest_value[var];
	    var_rest_value[var] = NONE;
	    push(var, VARIABLE_STACK);
	    if (var_current_value[var]==FALSE) {
		simple_manage_clauses(pos_in[var]);
		remove_clauses(neg_in[var]);
	    }
	    else {
		simple_manage_clauses(neg_in[var]);
		remove_clauses(pos_in[var]);
	    }
	    return FALSE;
	}
    } while (VARIABLE_STACK_fill_pointer != 0);
    return FALSE;
}

/*
int backtrack_to_root() {
   int var, index;
      
   UNITCLAUSE_STACK_fill_pointer = 0;
   do {
      var = pop(VARIABLE_STACK);
      var_state[var] = ACTIVE;
      for (index = saved_clause_stack[var]; 
               index < CLAUSE_STACK_fill_pointer;
               index++)
          clause_state[CLAUSE_STACK[index]] = ACTIVE;
      CLAUSE_STACK_fill_pointer = saved_clause_stack[var];

      for (index = saved_managedclause_stack[var];
               index < MANAGEDCLAUSE_STACK_fill_pointer;
               index++)
          clause_length[MANAGEDCLAUSE_STACK[index]]++;
      MANAGEDCLAUSE_STACK_fill_pointer = saved_managedclause_stack[var];
   } while (VARIABLE_STACK_fill_pointer != 0);
   return FALSE;
}
*/


int smaller_than(int lit1, int lit2) {
    return ((lit1<NB_VAR) ? lit1 : lit1-NB_VAR) < ((lit2<NB_VAR) ? lit2 : lit2-NB_VAR);
}

my_type redundant(int *new_clause, int *old_clause) {
    int i, j1=0, j2=0, lit1, lit2, old_clause_diff=0, new_clause_diff=0;
    
    while (((lit1=old_clause[j1]) != NONE) && ((lit2=new_clause[j2]) != NONE)) {
       if (smaller_than(lit1, lit2)) {
          j1++; old_clause_diff++;
       }
       else
       if (smaller_than(lit2, lit1)) {
          j2++; new_clause_diff++;
       }
       else
       if (complement(lit1, lit2)) {
           return FALSE; /* old_clause_diff++; new_clause_diff++; j1++; j2++; */
       }
       else {
          j1++; j2++;
       }
    }
    if ((lit1 == NONE) && (old_clause_diff == 0))
   /* la nouvelle clause est redondante ou subsumee */
       return NEW_CLAUSE_REDUNDANT;
    if ((lit2 == NONE) && (new_clause_diff == 0))
         /* la old clause est redondante ou subsumee */
       return OLD_CLAUSE_REDUNDANT;
    return FALSE;
}

my_type get_resolvant(int *clause1, int *clause2, int *resolvant) {
    int lit1, lit2, nb_diff=0, nb_iden=0, nb_opps=0, j1=0, j2=0, j, limited_length;

    while (((lit1=clause1[j1])!=NONE) && ((lit2=clause2[j2]) != NONE)) {
       if (complement(lit1, lit2)) {
          j1++; j2++; nb_opps++;
       }
       else
       if (lit1 == lit2) {
          j1++; j2++; nb_iden++;
       }
       else
       if (smaller_than(lit1, lit2)) {
          nb_diff++; j1++;
       }
       else {
          nb_diff++; j2++;
       }
    }
    if (nb_opps ==1) {
       if (clause1[j1] ==NONE) {
          for (; clause2[j2]!= NONE; j2++) nb_diff++;
       }
       else {
          for (; clause1[j1]!= NONE; j1++) nb_diff++;
       }
       if ((j1==1) || (j2==1))  limited_length=RESOLVANT_LENGTH; 
       else
       if ((j1==2) && (j2==2))  limited_length=1;
       else
       if (j1<j2) limited_length=j1;
       else  limited_length=j2;

       if (nb_diff + nb_iden <= limited_length) {
          j1=0; j2=0; j=0;
          while (((lit1 = clause1[j1])!=NONE) && ((lit2 = clause2[j2]) != NONE)) {
             if (lit1 == lit2) {
                resolvant[j] = lit1; j1++; j2++; j++;
             }
             else 
             if (smaller_than(lit1, lit2)) {
                resolvant[j] = lit1; j1++; j++;
             }
             else
             if (smaller_than(lit2, lit1)) {
                resolvant[j] = lit2; j2++; j++;
             }
             else {
                j1++; j2++;
             }
          }
          if (clause1[j1] ==NONE) while ((resolvant[j++] = clause2[j2++]) != NONE);
          else while ((resolvant[j++] = clause1[j1++]) != NONE);
          return TRUE;
       }
    }
    return FALSE;
}

void remove_link(int clause) {
   int lit;
   int *lits;
   struct node *pnode1, *pnode2, *pnode;

   lits = sat[clause];

   for (lit=*lits; lit != NONE; lit=*(++lits)) { 
       pnode = (positive(lit) ? node_pos_in[lit] : 
                                node_neg_in[get_var_from_lit(lit)]);
       if (pnode == NULL) return;
       if (pnode->clause == clause) {
          if (positive(lit)) node_pos_in[lit] = pnode->next;
          else node_neg_in[get_var_from_lit(lit)] = pnode->next;
       }
       else
       for (pnode1 = pnode, pnode2 = pnode->next;
            pnode2 != NULL; pnode2=pnode2->next) {
         if (pnode2->clause == clause) {
            pnode1->next = pnode2->next;
            break;
         }
         else
            pnode1 = pnode2;
       }
   }
}


int search_redundence(int *lits) {
   int i, lit, is_red;
   int *old_lits, *new_lits;
   struct node *pnode;

   new_lits = lits;
   for (lit=*lits; lit != NONE; lit=*(++lits)) { 
     for (pnode = (positive(lit) ? node_pos_in[lit] : 
                                   node_neg_in[get_var_from_lit(lit)]);
          pnode != NULL; pnode = pnode->next) {
       old_lits = sat[pnode->clause];
       is_red = redundant(new_lits, old_lits);
       if (is_red == OLD_CLAUSE_REDUNDANT) {
/*          printf("old clause %d is redundant\n", 
                 pnode->clause);
*/
          clause_state[pnode->clause] = PASSIVE;
          remove_link(pnode->clause);
       }
       else
       if (is_red == NEW_CLAUSE_REDUNDANT) {
          return NEW_CLAUSE_REDUNDANT;
       }
     }
  }
  return FALSE;  
}

int add_resolvant(int *lits) {
   int i, j, lit, is_red, resolvant[RESOLVANT_LENGTH+1];
   my_type is_res;
   int *old_lits, *new_lits, *res;
   struct node *pnode;
  
   /* return TRUE; */

   new_lits = lits;
   for (lit=*lits; lit != NONE; lit=*(++lits))
     for (pnode = (positive(lit) ?
                  node_neg_in[lit] :
                  node_pos_in[get_var_from_lit(lit)]);
          pnode != NULL; pnode = pnode->next) {
        old_lits = sat[pnode->clause];
        is_res = get_resolvant(new_lits, old_lits, resolvant);
        if (is_res != FALSE) {
          is_red=search_redundence(resolvant);
          if (is_red != NEW_CLAUSE_REDUNDANT) {
             res=(int *)malloc((RESOLVANT_LENGTH+1)*sizeof(int));
             clause_state[NB_CLAUSE]=ACTIVE;      
             j=0;
             while ((res[j]=resolvant[j]) != NONE) ++j;
             if (j==0) return NONE;
             sat[NB_CLAUSE] = res; 
             clause_length[NB_CLAUSE++]=j;
          }
        }
      }
   return TRUE;
}
   
my_type build_sat_instance(int numvar, int numclause, int * wff){

   char ch, tautologie, word1[WORD_LENGTH], word2[WORD_LENGTH];
   int endln, germe, i, j, index, length, ii, jj, lit1, is_red,
       lits[tab_variable_size], *plit, *lits1, lit, var, *clause,
       NB_CLAUSE1;
   int d;


   if (numvar > tab_variable_size || 
       numclause > tab_clause_size) return TOOBIG;

  NB_VAR = numvar;
  NB_CLAUSE = numclause;
  d=0;

   for (i=0; i<NB_CLAUSE; i++) {
      length=0; 
      lits[length] = wff[d++];
      while (lits[length] != 0) {
        length++;
	lits[length] = wff[d++];
      }
      tautologie = FALSE;
      /* test if some literals are redundant and sort the clause */
      for (ii=0; ii<length-1; ii++) {
         lit = lits[ii];
         for (jj=ii+1; jj<length; jj++) {
            if (abs(lit)>abs(lits[jj])) {
               lit1=lits[jj]; lits[jj]=lit; lit=lit1;
            }
            else
            if (lit == lits[jj]) {
               lits[jj--] = lits[--length]; lits[length] = 0;
               fprintf(STDMSG, "literal %d is redundant in clause %d\n", lit, i);
            }
            else
            if (abs(lit) == lits[jj]) {
               tautologie = TRUE; break;
            }
         }
         if (tautologie == TRUE) break;
         else lits[ii] = lit;
      }              
      if (tautologie == FALSE) {     
        lits[length] = 0;
        sat[i] = (int *)malloc((length+1)*sizeof(int));
        j=0;
        while (lits[j] != 0) { 
           if (lits[j] > 0) sat[i][j] = lits[j]-1;
           else sat[i][j] = abs(lits[j]) + NB_VAR -1;
           j++;
        }
        sat[i][j] = NONE;
        clause_length[i] = length;
        clause_state[i] = ACTIVE;
 /*       static_clause_length[i] = length;
*/
      }
      else i--;
   }

   for (i=0; i<NB_VAR; i++) {
      node_neg_in[i] = NULL;
      node_pos_in[i] = NULL;
      var_state[i]=ACTIVE;
   }

   INIT_NB_CLAUSE= NB_CLAUSE;
  
   for(i=0; i<NB_CLAUSE; i++) {
      plit = sat[i];
      length = clause_length[i];          
      if (length==1) push(i, UNITCLAUSE_STACK);
         if (search_redundence(plit) != NEW_CLAUSE_REDUNDANT) {
         if (add_resolvant(plit) == NONE) return NONE;
         else set_link(i);
      }
      else clause_state[i] = PASSIVE;
   }
   NB_CLAUSE1 = 0;
   for (i=0; i<NB_CLAUSE; i++) {
     if (clause_state[i] == ACTIVE) NB_CLAUSE1++;
   }
   NB_CLAUSE = NB_CLAUSE1;

   for(i=0; i<NB_VAR; i++) {
      neg_in[i]=copy_clauses(node_neg_in[i]);
      pos_in[i]=copy_clauses(node_pos_in[i]);
   }
   free_ressources();

   fprintf(STDMSG, "Wff loaded\n");

   return TRUE;
}


int verify_solution() {
   int i, j, lit, *lits, clause_truth;

   for (i=0; i<NB_CLAUSE; i++) {
      clause_truth = FALSE;
      lits = sat[i];
      for(lit=*lits; lit!=NONE; lit=*(++lits))
         if (((negative(lit)) && 
              (var_current_value[get_var_from_lit(lit)] == FALSE)) ||
             ((positive(lit)) && 
              (var_current_value[lit] == TRUE)) ) {
            clause_truth = TRUE;
            break;
         }
      if (clause_truth == FALSE) return FALSE;
   }

   return TRUE;
}

  
int unitclause_process() {
  int var, i, j, unitclause, clause, lit, *lits, unitclause_position;
  
  for (unitclause_position = 0; 
       unitclause_position < UNITCLAUSE_STACK_fill_pointer;
       unitclause_position++) {
     unitclause = UNITCLAUSE_STACK[unitclause_position];
     clause_length[unitclause]++;
     if (clause_state[unitclause] == ACTIVE) {
       NB_UNIT++;
       lits = sat[unitclause];
       for(lit=*lits; lit!=NONE; lit=*(++lits)) {
          if (positive(lit)) {
             var = lit;
             if (var_state[var] == ACTIVE) {
                var_current_value[var] = TRUE;
                var_rest_value[var] = NONE;
                if (manage_clauses(neg_in[var])==TRUE) {
                   var_state[var] = PASSIVE;
                   push(var, VARIABLE_STACK);
                   remove_clauses(pos_in[var]);
                   break;
                }
                else {
                   for (i=unitclause_position+1;  
                        i<UNITCLAUSE_STACK_fill_pointer; i++)
                      clause_length[UNITCLAUSE_STACK[i]]++;
		   return backtracking();
                }
             }
          }
          else {
             var = get_var_from_lit(lit);
             if (var_state[var] == ACTIVE) {
                var_current_value[var] = FALSE;
                var_rest_value[var] = NONE;
                if (manage_clauses(pos_in[var])== TRUE) {
                   var_state[var] = PASSIVE;
                   push(var, VARIABLE_STACK);
                   remove_clauses(neg_in[var]);
                    break;
                }
                else {
                   for (i=unitclause_position+1;  
                        i<UNITCLAUSE_STACK_fill_pointer; i++)
                      clause_length[UNITCLAUSE_STACK[i]]++;
		   return backtracking();
                }
             }
          }
       }     
     }
   }
   UNITCLAUSE_STACK_fill_pointer = 0;
   return FALSE;
}

int get_nb_clauses(int var) {
   return ((nb_neg_clause_of_length2[var] + 
            nb_pos_clause_of_length2[var]) * WEIGTH) + 
           nb_neg_clause_of_length3[var] + 
           nb_pos_clause_of_length3[var];
}

int get_resolvant_nb(int clause) {
  int *lits;
  int lit, var, i, resolvant_nb=0;
       
  lits = sat[clause];
  for(lit=*lits; lit!=NONE; lit=*(++lits)) {
     if (positive(lit)) {
        var = lit;
        if (var_state[var] == ACTIVE)
          resolvant_nb += (nb_neg_clause_of_length2[var] * WEIGTH)+ 
                           nb_neg_clause_of_length3[var];
     }
     else {
        var = get_var_from_lit(lit);
        if (var_state[var] == ACTIVE) 
           resolvant_nb += (nb_pos_clause_of_length2[var] * WEIGTH)+ 
                           nb_pos_clause_of_length3[var];
     }
  }
  return resolvant_nb;
}

int examine1(int tested_var) {
   int i, generating_fixed_variables_if_positif = 0, 
          generating_fixed_variables_if_negatif = 0,
          nb_clause_if_positif, nb_clause_if_negatif,
          clause, var;
   
      var_current_value[tested_var] = TRUE;

      var_state[tested_var] = PASSIVE;
      push(tested_var, MY_VARIABLE_STACK);
      my_simple_manage_clauses(neg_in[tested_var]);

      generating_fixed_variables_if_positif = branch();
   
      if (generating_fixed_variables_if_positif == NONE) {
         reset_context();
         var_current_value[tested_var] = FALSE;
         var_rest_value[tested_var] = NONE;
         var_state[tested_var] = PASSIVE;
         push(tested_var, VARIABLE_STACK);
         simple_manage_clauses(pos_in[tested_var]);
         remove_clauses(neg_in[tested_var]); 
         return NONE;
      }
      else {
         for (i=0; i<MY_MANAGEDCLAUSE_STACK_fill_pointer; i++) {
            clause = MY_MANAGEDCLAUSE_STACK[i];
            if (clause_length[clause] == 2)
               reduce_if_positive_nb[tested_var] += get_resolvant_nb(clause);
         }
         reset_context();
      }
      var_current_value[tested_var] = FALSE;

      var_state[tested_var] = PASSIVE;
      push(tested_var, MY_VARIABLE_STACK);
      my_simple_manage_clauses(pos_in[tested_var]);

      generating_fixed_variables_if_negatif = branch();

      if (generating_fixed_variables_if_negatif == NONE) {
         reset_context();
         simple_manage_clauses(neg_in[tested_var]);
         var_current_value[tested_var] = TRUE;
         var_rest_value[tested_var] = NONE;
         var_state[tested_var] = PASSIVE;
         push(tested_var, VARIABLE_STACK);
         remove_clauses(pos_in[tested_var]);
         return NONE;
       }
       else {
         for (i=0; i<MY_MANAGEDCLAUSE_STACK_fill_pointer; i++) {
            clause = MY_MANAGEDCLAUSE_STACK[i];
            if (clause_length[clause] == 2)
               reduce_if_negative_nb[tested_var] += get_resolvant_nb(clause);
         }
         reset_context();
      }
   push(tested_var, TESTED_VAR_STACK);
   return TRUE;
}

void manage_clauses_for_propagate(register int *clauses) {
   int clause;
   for(clause=*clauses; clause!=NONE; clause=*(++clauses)) {
      if (clause_state[clause] == ACTIVE) {
         clause_length[clause]--;
         push(clause, MANAGEDCLAUSE_STACK);
      }
   }
}

void simple_propagate(int lit) {
   int var;
   if (positive(lit)) {
      var=lit;
      var_current_value[var] = TRUE;
      manage_clauses_for_propagate(neg_in[var]);
      var_rest_value[var] = NONE;
      var_state[var]=PASSIVE;
      push(var, VARIABLE_STACK);
      remove_clauses(pos_in[var]);
   }
   else {
      var=get_var_from_lit(lit);     
      var_current_value[var] = FALSE;
      manage_clauses_for_propagate(pos_in[var]);
      var_rest_value[var] = NONE;
      var_state[var]=PASSIVE;
      push(var, VARIABLE_STACK);
      remove_clauses(neg_in[var]);
   }
}

int examine(int tested_var) {
   int i, generating_fixed_variables_if_positif = 0, 
          generating_fixed_variables_if_negatif = 0,
          nb_clause_if_positif,
          nb_clause_if_negatif,
          clause, 
          var, implied_lit;
   
   IMPLIED_LITS_fill_pointer=0;   

      var_current_value[tested_var] = TRUE;

      var_state[tested_var] = PASSIVE;
      push(tested_var, MY_VARIABLE_STACK);
      my_simple_manage_clauses(neg_in[tested_var]);

      generating_fixed_variables_if_positif = branch();
      reduce_if_positive_nb[tested_var]=MY_MANAGEDCLAUSE_STACK_fill_pointer;
   
      if (generating_fixed_variables_if_positif == NONE) {
         reset_context();
         var_current_value[tested_var] = FALSE;
         var_rest_value[tested_var] = NONE;
         var_state[tested_var] = PASSIVE;
         push(tested_var, VARIABLE_STACK);
         simple_manage_clauses(pos_in[tested_var]);
         remove_clauses(neg_in[tested_var]); 
         return NONE;
      }
      else {
         for (i=0; i<MY_VARIABLE_STACK_fill_pointer; i++) {  
            var=MY_VARIABLE_STACK[i];
            if (var_current_value[var]==TRUE) implied_lit=var;
            else implied_lit=var+NB_VAR;
            push(implied_lit, IMPLIED_LITS);
         }
         reset_context();
      }
      var_current_value[tested_var] = FALSE;

      var_state[tested_var] = PASSIVE;
      push(tested_var, MY_VARIABLE_STACK);
      my_simple_manage_clauses(pos_in[tested_var]);

      generating_fixed_variables_if_negatif = branch();
      reduce_if_negative_nb[tested_var]=MY_MANAGEDCLAUSE_STACK_fill_pointer;
      reset_context();

      if (generating_fixed_variables_if_negatif == NONE) {
 /*        simple_manage_clauses(neg_in[tested_var]);
         var_current_value[tested_var] = TRUE;
         var_rest_value[tested_var] = NONE;
         var_state[tested_var] = PASSIVE;
         push(tested_var, VARIABLE_STACK);
         remove_clauses(pos_in[tested_var]);
*/
         for (i=0; i<IMPLIED_LITS_fill_pointer; i++)
             simple_propagate(IMPLIED_LITS[i]);
         return NONE;
       }
   push(tested_var, TESTED_VAR_STACK);
   return TRUE;
}

int branch() {
   int var, lit, *lits, unitclause, clause, unitclause_position, i;
  
   NB_SEARCH++;
  
   for (unitclause_position = 0;
        unitclause_position != MY_UNITCLAUSE_STACK_fill_pointer;
        unitclause_position++) {
       unitclause = MY_UNITCLAUSE_STACK[unitclause_position];
       clause_length[unitclause]++;
          lits = sat[unitclause];
          for(lit=*lits; lit!=NONE; lit=*(++lits)) {
              if (positive(lit)) {
                 var = lit;
                 if (var_state[var] == ACTIVE) {
                    var_current_value[var] = TRUE;
                    if (my_manage_clauses(neg_in[var]) == TRUE) {
                       var_state[var] = PASSIVE;
                       push(var, MY_VARIABLE_STACK);
                       break;
                    }
                    else {
                       for (i= unitclause_position+1; 
                            i<MY_UNITCLAUSE_STACK_fill_pointer;
                            i++)
                           clause_length[MY_UNITCLAUSE_STACK[i]]++;
                       NB_FIXED++;
                       return NONE;
                    }
                 }
              }
              else {
                 var = get_var_from_lit(lit);
                 if (var_state[var] == ACTIVE) {
                    var_current_value[var] = FALSE;
                    if (my_manage_clauses(pos_in[var]) == TRUE) {
                       var_state[var] = PASSIVE;
                       push(var, MY_VARIABLE_STACK);
                       break;
                    }
                    else { 
                       for (i= unitclause_position+1;
                            i<MY_UNITCLAUSE_STACK_fill_pointer;
                            i++)
                           clause_length[MY_UNITCLAUSE_STACK[i]]++;
                       NB_FIXED++;
                       return NONE;
                    }
                 }
              }
             
          }
    }
   return TRUE;
}

int get_neg_clause_nb(int var) {
    my_type neg_clause3_nb = 0, neg_clause2_nb = 0;
    int *clauses, clause;
    clauses = neg_in[var];

    for(clause=*clauses; clause!=NONE; clause=*(++clauses)) {

/*    for (index = 0, end_value = neg_nb[var]; index<end_value; index++) {
       clause = clauses[index];
*/
       if (clause_state[clause] == ACTIVE) {
            if (clause_length[clause] == 2)
                neg_clause2_nb++;
            else
                neg_clause3_nb++;
        }
    }
    nb_neg_clause_of_length2[var] = neg_clause2_nb;
    nb_neg_clause_of_length3[var] = neg_clause3_nb;
    return neg_clause2_nb + neg_clause3_nb;
}

int get_pos_clause_nb(int var) {

    my_type pos_clause3_nb = 0, pos_clause2_nb = 0;
    int *clauses, clause;
    clauses = pos_in[var];

    for(clause=*clauses; clause!=NONE; clause=*(++clauses)) {

/*    for (index=0, end_value=pos_nb[var]; index<end_value; index++) {
        clause = clauses[index]; */

        if (clause_state[clause] == ACTIVE) {
            if (clause_length[clause] == 2)
                pos_clause2_nb++;
            else
                pos_clause3_nb++;
        }
    }
    nb_pos_clause_of_length2[var] = pos_clause2_nb;
    nb_pos_clause_of_length3[var] = pos_clause3_nb;
    return pos_clause2_nb + pos_clause3_nb;
}

long total_choices = 0;
double n = 0.0;
double sum_n = 0.0;
double sum_n_squared = 0.0;
double mean_n;
double variance_n;
double std_dev_n;
double std_error_mean_n;

int choose_and_instantiate_variable_in_clause() {
    int var, chosen_var=NONE; 
    int  i, j, posi, nega;
    int nb_clauses, max_nb_clauses = -1; 
    my_type pos2, neg2;
    long saved_nb_back;

    int num_good_vars = 0;
    
    NB_BRANCHE++;
    TESTED_VAR_STACK_fill_pointer=0;

    for (var = 0; var < NB_VAR; var++) {
       if (var_state[var] == ACTIVE) {
           reduce_if_negative_nb[var]=0;
           reduce_if_positive_nb[var]=0;

           if (get_neg_clause_nb(var) == 0) {
	       NB_MONO++;
               var_current_value[var] = TRUE;
               var_rest_value[var] = NONE;
               var_state[var] = PASSIVE;
               push(var, VARIABLE_STACK);
               remove_clauses(pos_in[var]);
           }
           else
           if (get_pos_clause_nb(var) == 0) {
               NB_MONO++;
               var_current_value[var] = FALSE;
               var_rest_value[var] = NONE;
               var_state[var] = PASSIVE;
               push(var, VARIABLE_STACK);
               remove_clauses(neg_in[var]);
           }
           else {
              pos2 = nb_pos_clause_of_length2[var];
              neg2 = nb_neg_clause_of_length2[var];
              if ((neg2>0) && (pos2>0) && ((neg2+pos2)>3) ) {
                 if (examine(var) == NONE) {
                    saved_nb_back = NB_BACK;
                    if (unitclause_process()==NONE) return NONE;
                    if (NB_BACK > saved_nb_back) return FALSE;
                 }
              }
           }
       }
    }

   if (TESTED_VAR_STACK_fill_pointer < T) {
   /*   TESTED_VAR_STACK_fill_pointer = 0; */
      for (var=0; var<NB_VAR; var++)
         if (var_state[var] == ACTIVE) {
            pos2 = nb_pos_clause_of_length2[var];
            neg2 = nb_neg_clause_of_length2[var];
            if ((neg2>0) && (pos2 > 0) &&
                ((neg2 > 1) || (pos2 > 1) ) )
               if (reduce_if_negative_nb[var]==0)
	          if (examine(var) == NONE) {
                     saved_nb_back = NB_BACK;
                     if (unitclause_process()==NONE) return NONE;
                     if (NB_BACK > saved_nb_back) return FALSE;
                  }
         }
   }

   if (TESTED_VAR_STACK_fill_pointer < T) {
      TESTED_VAR_STACK_fill_pointer = 0;
      for (var=0; var<NB_VAR; var++)
         if (var_state[var] == ACTIVE)
	     if (examine1(var) == NONE) {
                saved_nb_back = NB_BACK;
                if (unitclause_process()==NONE) return NONE;
                if (NB_BACK > saved_nb_back) return FALSE;
             }
   }
   num_good_vars = 0;
   /* Bart's version - works well, but I cannot understand it!
   for (i=0; i<TESTED_VAR_STACK_fill_pointer; i++) {
       var=TESTED_VAR_STACK[i];
       if (var_state[var] == ACTIVE) {
          posi=reduce_if_positive_nb[var];
          nega=reduce_if_negative_nb[var];
          nb_clauses = posi*nega*1024 + posi + nega +1;
          if ((nb_clauses >= max_nb_clauses) &&
              ((float)nb_clauses <= 
               ((1.0 + factor) * (float) max_nb_clauses))) {
             good_vars[num_good_vars] = var;
             num_good_vars++;
          }
          if ((float)nb_clauses > 
              ((1.0 + factor) * (float) max_nb_clauses)) {
             good_vars[0] = var;
             num_good_vars = 1;
             max_nb_clauses = nb_clauses;
          }
        }
   } 
   */


   /* My version - sensible but will it work? */
   for (i=0; i<TESTED_VAR_STACK_fill_pointer; i++) {
       var=TESTED_VAR_STACK[i];
       if (var_state[var] == ACTIVE) {
          posi=reduce_if_positive_nb[var];
          nega=reduce_if_negative_nb[var];
          nb_clauses = posi*nega*1024 + posi + nega +1;
          if (nb_clauses >= max_nb_clauses) 
	      max_nb_clauses = nb_clauses;
       }
   }
   for (i=0; i<TESTED_VAR_STACK_fill_pointer; i++) {
       var=TESTED_VAR_STACK[i];
       if (var_state[var] == ACTIVE) {
          posi=reduce_if_positive_nb[var];
          nega=reduce_if_negative_nb[var];
          nb_clauses = posi*nega*1024 + posi + nega +1;
	  if (max_nb_clauses == 0 ||
	      ((float)nb_clauses)/((float)max_nb_clauses) >= (1.0 - factor)){
	      good_vars[num_good_vars] = var;
	      num_good_vars++;
          }
       }
   }
   /* end my version */

   if (num_good_vars > 0) {
       chosen_var = good_vars[random()%num_good_vars];
       total_choices ++;
       sum_n += num_good_vars;
       sum_n_squared += num_good_vars * num_good_vars;
       mean_n = sum_n / total_choices;
   }
   if (chosen_var == NONE) return FALSE;
   /* if (chosen_var == NONE) return NONE; */

   var_state[chosen_var] = PASSIVE;
   saved_clause_stack[chosen_var] = CLAUSE_STACK_fill_pointer;
   saved_managedclause_stack[chosen_var] = MANAGEDCLAUSE_STACK_fill_pointer;
   push(chosen_var, VARIABLE_STACK);
   var_current_value[chosen_var] = TRUE;
   var_rest_value[chosen_var] = FALSE;
   simple_manage_clauses(neg_in[chosen_var]);
   remove_clauses(pos_in[chosen_var]);
   return FALSE;
}

my_type build(int simple_sat_instance, int numvar, int numclause, int * wff){
    return build_sat_instance(numvar, numclause, wff);
}

long back_this_solution;
long int totalback = 0;		/* total number of flips in all tries so far */
long int totalsuccessback = 0;	/* total number of flips in all tries which succeeded so far */
int numsuccesstry = 0;		/* total found solutions */





long x = 0;
long integer_sum_x = 0;
double sum_x = 0.0;
double sum_x_squared = 0.0;
double mean_x;
double second_moment_x;
double variance_x;
double std_dev_x;
double std_error_mean_x;
double seconds_per_back;
int r = 0;
int sum_r = 0;
double sum_r_squared = 0.0;
double mean_r;
double variance_r;
double std_dev_r;
double std_error_mean_r;

double s = 0.0;
double sum_s = 0.0;
double sum_s_squared = 0.0;
double mean_s;
double variance_s;
double std_dev_s;
double std_error_mean_s;


void update_statistics_end_restart(int successful)
{
    int i;
    int j;
    /* FILE *fp_time; */


    totalback += NB_BACK;
    x += NB_BACK;
    r ++;
    s = elapsed_seconds();
    sum_s += s;

    if (successful){
	numsuccesstry ++;
	sum_s_squared += s * s;
	mean_s = sum_s / numsuccesstry;
	totalsuccessback += NB_BACK;
	integer_sum_x += x;
	sum_x = (double) integer_sum_x;
	sum_x_squared += ((double)x)*((double)x);
	mean_x = sum_x / numsuccesstry;
	sum_r += r;
	mean_r = ((double)sum_r)/numsuccesstry;
	sum_r_squared += ((double)r)*((double)r);


	fprintf(STDMSG, "[%d] %s %5.3f %ld %ld %ld %ld %d %d %d %d %d %d %d\n", restart,
	       VERSION,
	       s,
	       NB_BRANCHE, NB_BACK,  NB_SEARCH, NB_FIXED, 
	       satisfiable(), NB_VAR, INIT_NB_CLAUSE, NB_CLAUSE-INIT_NB_CLAUSE,
	       cutoff, restart_max, seed);

	fflush(STDMSG);
	x = 0;
	r = 0;
	s = 0.0;
	(void) elapsed_seconds();
    }
    else {
	fprintf(STDMSG, "[%d] %s %5.3f %ld %ld %ld %ld %d %d %d %d %d %d %d\n", restart,
	       VERSION,
	       s,
	       NB_BRANCHE, NB_BACK,  NB_SEARCH, NB_FIXED, 
	       -1, NB_VAR, INIT_NB_CLAUSE, NB_CLAUSE-INIT_NB_CLAUSE,
	       cutoff, restart_max, seed);
    }

}

void print_statistics_final(void)
{
    double expertime;

    expertime = elapsed_seconds() + sum_s;
    seconds_per_back = expertime / totalback;
    fprintf(STDMSG, "\n");
    fprintf(STDMSG, "total elapsed seconds = %f\n", expertime);
    fprintf(STDMSG, "average backtracks per second = %d\n", (int)(totalback/expertime));
    fprintf(STDMSG, "number solutions found = %d\n", numsuccesstry);
    fprintf(STDMSG, "final success rate = %f\n", ((double)numsuccesstry * 100.0)/restart);
    fprintf(STDMSG, "average backtracks per successful restart = %li\n", numsuccesstry ? (totalsuccessback/numsuccesstry) : 0);
    if (numsuccesstry > 0)
	{
	    fprintf(STDMSG, "mean backtracks until solution = %f\n", mean_x);
	    if (numsuccesstry>1){
		second_moment_x = sum_x_squared / numsuccesstry;
		variance_x = second_moment_x - (mean_x * mean_x);
		/* Adjustment for small small sample size */
		variance_x = (variance_x * numsuccesstry)/(numsuccesstry - 1);
		std_dev_x = sqrt(variance_x);
		std_error_mean_x = std_dev_x / sqrt((double)numsuccesstry);
		fprintf(STDMSG, "  variance = %f\n", variance_x);
		fprintf(STDMSG, "  standard deviation = %f\n", std_dev_x);
		fprintf(STDMSG, "  standard error of mean = %f\n", std_error_mean_x);
	    }
	    fprintf(STDMSG, "mean seconds until solution = %f\n", mean_s);
	    if (numsuccesstry>1){
		variance_s = (sum_s_squared / numsuccesstry) - (mean_s * mean_s);
		variance_s = (variance_s * numsuccesstry)/(numsuccesstry - 1);	   
		std_dev_s = sqrt(variance_s);
		std_error_mean_s = std_dev_s / sqrt((double)numsuccesstry);
		fprintf(STDMSG, "  variance = %f\n", variance_s);
		fprintf(STDMSG, "  standard deviation = %f\n", std_dev_s);
		fprintf(STDMSG, "  standard error of mean = %f\n", std_error_mean_s);
	    }
	    fprintf(STDMSG, "mean restarts until solution = %f\n", mean_r);
	    if (numsuccesstry>1){
		variance_r = (sum_r_squared / numsuccesstry) - (mean_r * mean_r);
		variance_r = (variance_r * numsuccesstry)/(numsuccesstry - 1);	   
		std_dev_r = sqrt(variance_r);
		std_error_mean_r = std_dev_r / sqrt((double)numsuccesstry);
		fprintf(STDMSG, "  variance = %f\n", variance_r);
		fprintf(STDMSG, "  standard deviation = %f\n", std_dev_r);
		fprintf(STDMSG, "  standard error of mean = %f\n", std_error_mean_r);
	    }
	}

    fprintf(STDMSG, "mean choices each branchpoint (noise level) = %f\n", mean_n);
    if (total_choices>1){
	variance_n = (sum_n_squared / total_choices) - (mean_n * mean_n);
	std_dev_n = sqrt(variance_n);
	std_error_mean_n = std_dev_n / sqrt((double)total_choices);
	fprintf(STDMSG, "  variance = %f\n", variance_n);
	fprintf(STDMSG, "  standard deviation = %f\n", std_dev_n);
	fprintf(STDMSG, "  standard error of mean = %f\n", std_error_mean_n);
    }
}


int bb_satsolve_satz( int numvar, int numclause, int * wff, 
			int * soln, int argc, char * argv[] )
{
    clock_t begintime, endtime;
    int i,  begin, end, mess; 
    struct tms *a_tms;
    int a;
    int simple_sat_instance = 0;
    int restarted;
    int numsol;
    int nosol = TRUE;
    int isKnownSat = FALSE;
    int isKnownUnsat = FALSE;

    int answer;






    /* static vars, manually initialize*/
    NB_SEARCH = 0; NB_FIXED = 0;
    CLAUSE_NODES_pointer=MAX_NODE_NUMBER;
    VARIABLE_STACK_fill_pointer = 0;
    CLAUSE_STACK_fill_pointer = 0;
    UNITCLAUSE_STACK_fill_pointer = 0;
    MANAGEDCLAUSE_STACK_fill_pointer = 0;
    IMPLIED_LITS_fill_pointer=0;
    TESTED_VAR_STACK_fill_pointer=0;
    R = 3;
    NB_UNIT=0; NB_MONO=0; NB_BRANCHE=0; NB_BACK = 0;
    MAYBE_SATISFIABLE = TRUE;
    MY_VARIABLE_STACK_fill_pointer = 0;
    MY_UNITCLAUSE_STACK_fill_pointer = 0;
    MY_MANAGEDCLAUSE_STACK_fill_pointer = 0;
    factor = -1.0;
    cutoff = 1000000000;
    restart = 0;
    restart_max = 1;
    total_choices = 0;
    n = 0.0;
    sum_n = 0.0;
    sum_n_squared = 0.0;
    totalback = 0;
    totalsuccessback = 0;
    numsuccesstry = 0;
    x = 0;
    integer_sum_x = 0;
    sum_x = 0.0;
    sum_x_squared = 0.0;
    r = 0;
    sum_r = 0;
    sum_r_squared = 0.0;
    s = 0.0;
    sum_s = 0.0;
    sum_s_squared = 0.0;

    fprintf(STDMSG, "Invoking solver satz version %s\n", VERSION);

    gettimeofday(&tv,&tzp);
    seed = (( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec;

    numsol = 1;
    a=0;
    while (++a < argc){
	if (strcmp(argv[a],"-seed")==0)
	    seed = atoi(argv[++a]);
	else if (strcmp(argv[a],"-cutoff")==0)
	    cutoff = atoi(argv[++a]);
	else if (strcmp(argv[a],"-numsol")==0)
	    numsol = atoi(argv[++a]);
	else if (strcmp(argv[a],"-noise")==0)
	    factor = atof(argv[++a]);
	else if (strcmp(argv[a],"-restart")==0)
	    restart_max = atoi(argv[++a]);
	else if (strcmp(argv[a],"-printsol")==0)
	    nosol = FALSE;
	else {
	    fprintf(STDMSG, "Options for satz (satz-rand) solver are:\n"
		    "    -seed N = initialize random seed to N\n"
		    "    -cutoff N = stop search after N backtracks, and then (possibly) restart\n"
		    "    -restart N = restart up to N times or until problem is solved\n"
		    "    -numsol N = keep going until the problem is solved N times (default 1)\n"
		    "    -noise F = degrade variable choice heuristic to increase randomness\n"
		    "                  between restarts: 1.0 = random, 0.0 = no noise,\n"
		    "                  default = 0.4 if restart > 1, otherwise default = 0.0\n");
	    exit(1);
	}
    }

    /* Should the factor be 0 if there are no restarts? 
       Seems to HURT on some wffs! */
    if (factor < 0.0){
	if (restart_max == 1) factor = 0.0; /* Maybe make factor = 0.0 in this case */
	else factor = 0.4;
    }
    if (factor > 1.0) factor = 1.0;

    srandom(seed);
    fprintf(STDMSG, "seed: %d\ncutoff: %d\nnoise: %f\nrestart: %d\n", seed, cutoff, factor, restart_max);
    fprintf(STDMSG, "fields: time NB_BRANCH NB_BACK NB_SEARCH NB_FIXED sat NB_VAR INIT_NB_CLAUSE clauses_added cutoff restart seed\n");

    fflush(stdout);
    restart = 0;

    a_tms = ( struct tms *) malloc( sizeof (struct tms));
    mess=times(a_tms); begintime = a_tms->tms_utime;
    (void) elapsed_seconds();

    switch (build(simple_sat_instance, numvar, numclause, wff)){
    case TOOBIG:
	fprintf(STDMSG, "satz-rand: wff too big, must recompile!\n");
	return Failure;
    case FALSE: 
	fprintf(STDMSG, "satz-rand: Input format error\n"); 
	exit(1);
    case TRUE:
	save_everything();
	answer = Timeout;
	while (numsuccesstry < numsol && restart < restart_max){
	    restore_everything();
	    do {
		if (UNITCLAUSE_STACK_fill_pointer==0)
		    restarted = choose_and_instantiate_variable_in_clause();
		if (restarted != NONE) restarted = unitclause_process();
	    }  while (restarted != NONE && 
		      (restarted==TRUE ||( VARIABLE_STACK_fill_pointer != 0 &&  !satisfiable())));
	    if (restarted == NONE){
		if (restart >= restart_max)
		    fprintf(STDMSG, "**** reached maximum restarts ****\n");
		else
		    fprintf(STDMSG, "**** reached blackbox resource limits ****\n");
		update_statistics_end_restart(FALSE);
		break;
	    }
	    else if (satisfiable()){
		isKnownSat = TRUE;
		fprintf(STDMSG, "**** the instance is satisfiable *****\n");
		if (verify_solution()) {
		    fprintf(STDMSG, "**** verification of solution is OK ****\n");
		    update_statistics_end_restart(TRUE);
		    if (numsuccesstry == 1) save_solution(NB_VAR, soln);
		}
		else {
		    fprintf(STDMSG, "**** PROGRAM ERROR, VERIFICATION FAILS ****\n");
		    exit(-1);
		}
	    }
	    else {
		fprintf(STDMSG, "**** the instance is unsatisfiable *****\n");
		update_statistics_end_restart(TRUE);
		isKnownUnsat = TRUE;
	    }
	}

	break;
    case NONE: 
	fprintf(STDMSG, "**** an empty resolvant is found during preprocessing ****\n");
	fprintf(STDMSG, "**** the instance is unsatisfiable *****\n");
	answer = Unsat;
	break;
    }

    print_statistics_final();

    if (!nosol && isKnownSat) print_solution(NB_VAR);

    mess=times(a_tms); endtime = a_tms->tms_utime;
    fprintf(STDMSG, "Program terminated in %5.3f seconds.\n",
	    ((double)(endtime-begintime)/CLK_TCK));

    if (isKnownSat) answer = Sat;
    if (isKnownUnsat) answer = Unsat;

    return answer;
}
